#include <stdio.h>
#include <stdlib.h>



#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <sys/wait.h>

#include <string.h>

typedef struct {
  pid_t    pid;
  FILE   * pipe;
} sh_gpg_popen_t;


void sh_unix_closeall (int fd)
{

#ifdef OPEN_MAX
  int fdlimit = OPEN_MAX;
#else
  int fdlimit = sysconf(_SC_OPEN_MAX);
#endif

  /* can't happen - so fix it :-(
   */
  if (fdlimit < 0)
    fdlimit = 8;  /* POSIX lower limit */

  if (fdlimit > 65536)
    fdlimit = 65536;

  /* Close everything from fd (inclusive) up to fdlimit (exclusive). 
   */

  while (fd < fdlimit)
    close(fd++);

  return;
}


static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, char * filepath, 
			    int mode, char * id)
{
  int pipedes[2];
  FILE * outf;
  char * envp[2];
  char   path[256];
  char   cc1[32];
  char   cc2[32];

  char * arg[6];

  strncpy (path,  "/bin/ls", 256);
  strncpy (cc1,   "-ltr",     32);
  strncpy (cc2,   "-n",       32);


  arg[0] = path; 
  arg[1] = cc1;
  arg[3] = filepath;
  arg[4] = NULL;

  (void) mode;
  (void) id;

  envp[0] = malloc (strlen(getenv("HOME")) + 6);
  sprintf (envp[0], "HOME=%s", getenv("HOME"));  /* known to fit  */
  envp[1] = NULL;

  fprintf(stderr, "TP 1\n");

  if (pipe(pipedes) < 0) 
    {
      if (envp[0] != NULL) 
	free(envp[0]);
      return (NULL);
    }
  

  fprintf(stderr, "TP 2\n");

  source->pid = fork();

  /* Failure
   */
  if (source->pid == (pid_t) - 1) 
    {
      fprintf(stderr, "TP 3 - failure\n");

      close(pipedes[0]);
      close(pipedes[1]);
      if (envp[0] != NULL) 
	free(envp[0]);
      return (NULL);
    }

  if (source->pid == (pid_t) 0) 
    {

      fprintf(stderr, "CHILD 1\n");

      /* child - make read side of the pipe stdout 
       */
      if (dup2(pipedes[STDOUT_FILENO], STDOUT_FILENO) < 0)
	_exit(EXIT_FAILURE);
      
      fprintf(stderr, "CHILD 2\n");

      /* close the pipe descriptors 
       */
      close (pipedes[STDIN_FILENO]);
      close (pipedes[STDOUT_FILENO]);
      
      fprintf(stderr, "CHILD 3\n");

      /* don't leak file descriptors
       */
      sh_unix_closeall (3);

      fprintf(stderr, "CHILD 4\n");

      freopen("/dev/null", "r+", stderr); 
      
      fprintf(stderr, "CHILD 5\n");

      /* We should become privileged if SUID,
       * to be able to read the keyring.
       * We have checked that gpg is OK,
       * AND that only a trusted user could overwrite
       * gpg.
       */
      /* memset (skey, '\0', sizeof(sh_key_t)); */
      /* sl_set_suid(); */
      setuid(geteuid());
      
      /* exec the program */
      execve ("/bin/ls", arg, envp);
      
      /* failed 
       */
      _exit(80);
    }

  /* parent
   */

  fprintf(stderr, "PARENT 1\n");

  if (envp[0] != NULL) 
    free(envp[0]);


  fprintf(stderr, "PARENT 2\n");

  close (pipedes[STDOUT_FILENO]);
  fcntl (pipedes[STDIN_FILENO], F_SETFD, FD_CLOEXEC);


  fprintf(stderr, "PARENT 3\n");
  outf = fdopen (pipedes[STDIN_FILENO], "r");
  

  fprintf(stderr, "PARENT 4\n");
  if (outf == NULL) 
    {

      fprintf(stderr, "PARENT 4 - failed\n");
      kill (source->pid, SIGKILL);
      close (pipedes[STDOUT_FILENO]);
      waitpid (source->pid, NULL, 0);
      source->pid = 0;
      return (NULL);
    }
  
  fprintf(stderr, "PARENT 5\n");
  return (outf);
}



static int sh_gpg_pclose (sh_gpg_popen_t *source)
{
  int status = 0;
  
  fprintf(stderr, "CLOSE 1\n");
  if (fclose(source->pipe))
    return (-1);
  
  fprintf(stderr, "CLOSE 2\n");
  if (waitpid(source->pid, NULL, 0) != source->pid)
    status = -1;
  
  fprintf(stderr, "CLOSE 3\n");
  source->pipe = NULL;
  source->pid = 0;
  return (status);
}
 
static
int sh_gpg_check_file_sign(char * filepath, char * sign_id, char * sign_fp)
{

  char line[256];
  sh_gpg_popen_t  source;


 fprintf(stderr, "RUN 1\n");
 fflush(NULL); 
 source.pipe   = sh_gpg_popen  ( &source, filepath, 0, NULL );


  fprintf(stderr, "RUN 2\n");
  if (NULL == source.pipe)
    return 8;

  fprintf(stderr, "RUN 3\n");
  while (NULL != fgets(line, 255, source.pipe))
    {
      fprintf(stderr, "RECV: %s", line);
    }  
  fprintf(stderr, "RUN 5\n");
  sh_gpg_pclose (&source);

  fprintf(stderr, "RUN 6\n");
  return 0;
}

int main ()
{
  fprintf(stderr, "MAIN 1\n");
  sh_gpg_check_file_sign("/var", "xx", "yy");
  fprintf(stderr, "MAIN 2\n");

  return 0;
}
    




